#
#           main.py -- QCTLMC [francoisl@irif.fr]
#


import os
import sys
import time

from structure import *
import qctl  

from formule import *

def PrintCol(s,c) :
    #col="Style.BRIGHT+Fore."+c
    #print(eval(col)+s)
    #print(Style.RESET_ALL,end="")
    print(s)

sys.setrecursionlimit(1500)

VERBOSE=False
SOLVER='X'
AUTOTEST=False
TIMEOUT=True

# On a ici la procedure principale de model-checking.
# On lit notre structure et notre formule a partir d'un fichier,
# puis on les transforme en une formule QBF. On ecrit cette formule
# dans un fichier temporaire, puis on appelle Z3 dessus.

import ply.yacc as yacc
from qctl import parser as parser

def read_formula(pathname) :
    input=open(pathname, "r")
    s=input.read()
    input.close()
    F=parser.parse(s)
#    print(F.toString())
    return F

def read_structure(pathname) :
    K=Krip.from_file(pathname)
    return K

def parse_formula(s) :
    F=parser.parse(s)
#    print(F.toString())
    return F


def getF(s) :
    return parse_formula(s)


####################################################################################


def P(x) :
    print(x[0].toString())
    print((x[1]).toString())
    
# Reduction is used to transform the initial QCTL spec into another QCTL formula.
# for example: it may include the flatening of the formula, new quantifications, etc.
def Reduction(F,methode,solver ='qfm') :
    F=F.SimplifB()
    if (methode=="DIRECT") :
        (VarQ,VarQone,VarQoneForall) = F.PropQuant()
        return (F,RPARAM(VarQ,VarQone))
    if (solver=='cqesto' or solver=='ghostq' or solver=='qfun') and (methode != 'PNF1') and (methode != 'PNF2') and (methode != 'FBV') :
        PrintCol('Warning ! This solver requires the reduction PNF1.','RED')
#        return (BOTTOM(),None) 
    if methode=='UU' :     ########    UU
#        F=toNNF(F)
        if solver in ['Z3'] : 
            F=F.FNN(False,True)    # on laisse les EQ
        else :
            F=F.FNN(False,False)   # on les remplace
        (VarQ,VarQone,VarQoneForall) = F.PropQuant()
        if VERBOSE : 
            print("(size, temporal height, nb of alt, nb of qu. var) = ",F.Size())
            print("NB: for UU reduction, Untils are reduced in the QBF step")
        return (F,RPARAM(VarQ,VarQone))
    elif methode=='FUU' :  ########    FUU 
        F = F.FNN(False)
        F=F.SimplifB()
        F=F.MaPl({},{},-1,0)   # cette MaP renvoie la formule de la prop 9: ... AG (ki <=> Phi_i)
        (VarQ,VarQone,VarQoneForall) = F.PropQuant()
        if VERBOSE : 
            print("Flatening :",F.toString())
            print("Quantif. var :",VarQ)
            print("Quantif. var for Ex1/Al1 : ",VarQone)
            print("(size, temporal height, nb of alt, nb of qu. var) = ",F.Size())
            print("NB: for UU reduction, Untils are reduced in the QBF step")
        return (F,RPARAM(VarQ,VarQone))
    elif methode=='FP' :     ########    FP
        F=F.FPT()
        (VarQ,VarQone,VarQoneForall) = F.PropQuant()
        if VERBOSE :
            print("Quantif. var :",VarQ)
            print("Quantif. var for Ex1/Al1 : ",VarQone)
            print("(size, temporal height, nb of alt, nb of qu. var) = ",F.Size())
        return (F,RPARAM(VarQ,VarQone))
    elif methode=='FFP' :    ########    FFP
        F = F.FNN(False)
        F=F.SimplifB()
        F=F.MaPl({},{},-1,0)
        F=F.FPT()
        (VarQ,VarQone,VarQoneForall) = F.PropQuant()
        if VERBOSE :
            print("Quantif. var :",VarQ)
            print("Quantif. var for Ex1/Al1 : ",VarQone)
            print("(size, temporal height, nb of alt, nb of qu. var) = ",F.Size())
        return (F,RPARAM(VarQ,VarQone))
    elif methode=='FFNN' or methode=='FBV' or methode=='PNF1' or methode=='PNF2':    ########    FFNN  or FBV
        F=F.SimplifB()
        F = F.FNN(False,True)  # True car on laisse les EQ
        D,Df={ },{ }
        newproppf=None # for PNF1/PNF2 to encode pppf
        if methode=='FBV' :     # les deux MaP sont un chouille differentes
            F = F.MaPl(D,Df,0,1)
        elif methode=='PNF2' :
            F = F.MaPl(D,Df,0,3)  # on laisse les operateurs dans la formule de base (y compris EU, AU...)
            if F.ContainsOP([EU,AU,AF]) :
                newproppf=getIdent()
                print("(PNF2 reduction) The main formula contains lfp (AF/EU/AU). ",newproppf)
        else :     # FFNN or PNF1 
            F = F.MaPl(D,Df,0,2)  # on recupere la formule Phi_0 et les Theta_i que l'on va traiter ensuite
        Lnewprop = [D[index] for index in D]
        if (methode=='PNF1' or methode=='PNF2') and newproppf==None : # on verifie si il y a des pppf dans les theta_i (si  newproppf deja def, on sait deja qu'il faut ajouter le quantif Forall)
            for fname in Df :
                Faux = Df[fname]
                tofF = str(type(Faux))
                tofF=tofF[16:-2]
                if tofF in ['EU', 'AU', 'EF', 'AF'] :
                    newproppf=getIdent()  # une seule var suffit pour toutes les formules pppf
                    print("(PNF1-PNF2 reduction) The formula contains lfp. ",newproppf)
                    break
        if len(Lnewprop) > 0 or (newproppf != None): 
            F = F.AjoutPropQuant(Lnewprop,newproppf) # ajoute un "dernier" bloc Exists q1...qk avant la formule aplatie suivie eventuellement par Forall newproppf
        if methode=='PNF2' :
            F=F.TraitementEUAUAF(newproppf) 
        F=F.SimplifB() # utile ?
        if VERBOSE :
            print("Flattening : ",F.toString())
            for index in D :
                print(D[index]," -> ",index)
        (VarQ,VarQone,VarQoneForall) = F.PropQuant()  # on ne s'interesse pas a VarOneForall (uniquement pour BV et Forall1)
        # on distingue dans VarQ, les var correspondant a un EU/AU/EF/AF
        PropB, PropBVpppf = [ ], [ ] # on va distinguer les  prop. l (pour EW et AW) et les BitVec (pour EU et AU) pour FBV !
        # pour FFNN, tout est dans propB
#        Df2, Lconj = { } ,[ ]
        Lconj = [ ]
        for fname in Df : # on construit les reductions pour les differents cas: EW, AW,...
            # pour avoir AG (k_i -> theta_i)
            Faux = Df[fname]
            pFaux = D[fname]
#            Df2[fname] = Faux.QcirReduc(pFaux,VarQ)
#            Lconj.append(Df2[fname])
            if methode=='FBV' :  # les Until sont codes que dans le passage a QBF. On les laisse ici...
                tofF = str(type(Faux))
                tofF=tofF[16:-2]
                if tofF in ['EU', 'AU', 'EF', 'AF'] :            # ce sont des BitVec
                    PropBVpppf.append(pFaux)
                    Lconj.append(AG(IMPL(AP(pFaux),Faux)))
                else :      # ce sont des var boolennes
                 #   PropB.append(pFaux)
                    Lconj.append(Faux.QcirReduc(pFaux,VarQ,1))  # c'est la meme que pour qcir ici
            elif methode=='FFNN'  :    # FFNN ou PNF1
              #  PropB.append(pFaux)
                Lconj.append(Faux.QcirReduc(pFaux,VarQ,1))  # version=1
            else : # PNF1
                Lconj.append(Faux.QcirReduc(pFaux,VarQ,2,newproppf))  # version=2
        F= F.AjoutListConj(Lconj)
        if methode=='FBV' :   # on retire de VarQ les var qui seront des BV avec pppf
            aux=[]
            for v in VarQ :
                if not (v in PropBVpppf) :
                    aux.append(v)
            VarQ=aux
        if VERBOSE :
            print("Finally :")
            print(F.toString())
            print("VarQ = ",str(VarQ)," VarBVlfp = ",str(PropBVpppf)," VarQone = ",str(VarQone))
            print("(size, temporal height, nb of alt, nb of qu. var) = ",F.Size())
            if methode=='FBV' :
                print("NB: for FBV reduction, Untils are reduced in the QBF step")
        return (F,RPARAM(VarQ,VarQone,PropBVpppf))
    else :
        print("Unknown method.")
        return (TOP(),RPARAM())


    # construction d'une formule QBF pour F
    # dela depend de la methode, du caractere Prenex ou non, et du traitement des EA1
    # EA1 : 1 -> methode explicite, 2 -> methode par pt fixe, 3 -> methode des BV
def QBF(F,rp,krip,method,PNF,EA1) :
        s = krip.start
        Nbstates = len(krip.accessibles[s])
        # d'abord on insere des UNIQ() aux bons endroits:
        QBvar,QvarOne,QBVvarpf = rp.BQProp,rp.QPropEA, rp.BVQProppf
        if method != 'FBV' :
            assert(len(QBVvarpf)==0)
        luniq=[]
        Fu = F.AjoutUniq(PNF,luniq)
#        print("Luniq : ",str(luniq))
#        print('Uniq: ',Fu.toString())
        if VERBOSE :
                print("Inserting Uniq()... ",Fu.toString())
#        if len(QBVvarea) >0  :
#            print("Warning : the list of quantif. var for Ex1/Al1 in a spec should be empty here (QBF function) !")
#        for fu in luniq :
#            QBVvarea.append(fu.ap)
        k=0
        if method=='FBV' :  # on veut coder les val 0...N
            rp.Kpf=math.ceil(math.log(krip.BUntil+1,2))
        if EA1==3 :   # dans ce cas on veut coder les val 0...N-1
            rp.Kea=math.ceil(math.log(Nbstates,2))
            if VERBOSE :
                print("Size of BitVec to encode Ex1/Al1 :",rp.Kea)
        elif EA1==6 : # codage Ex1 comme "au plus un" via des BV a valeur entre 0 et N
            rp.Kea=math.ceil(math.log(Nbstates+1,2))
            if VERBOSE :
                print("Size of BitVec to encode Ex1/Al1 :",rp.Kea)
        rp.method=method
        rp.codageEA1 = EA1
#        if (method=='FBV' and SOLVER=='z3') :
#            print("Warning ! Reduction for Z3 uses built-in bitvectors !")
#            print("To be done !")
        #print("0-> "+Fu.toString())
        Fq=Fu.toQBF(rp,krip,s)
        #print("1-> "+Fq.toString())
        Fq=Fq.SimplifB()
        #print("2-> "+Fq.toString())
        return Fq

        
# ecrit dans fichname la formule QBF F au format QCIR
def toQCIR(F,fichname) :
    print("---- Translation to QCIR  ----------------------------")
    if os.path.isfile(fichname) :
        print("on supprime le vieux fichier out: ",fichname)
        if os.name != "nt" :
            os.system("rm "+fichname)
        else :
            os.system("del "+fichname)
    print("file: "+fichname)
    fout=open(fichname, 'w')
    tab = { }
    fout.write("#QCIR-G14 \n")
    aux = F.toQCIR(tab,fout,True,None)
    fout.close()

def toZ3(F,fichname) :
    print("---- Translation to Z3 format  ----------------------------")
    if os.path.isfile(fichname) :
        print("on supprime le vieux fichier out: ",fichname)
        if os.name != "nt" :
            os.system("rm "+fichname)
        else :
            os.system("del "+fichname)
    print("file: "+fichname)
    fout=open(fichname, 'w')
    fout.write("(assert ")
    aux = F.toZ3(fout)
    fout.write(")\n")
    fout.write("(check-sat)\n")
    print("end of translation to Z3")
    fout.close()

def QBFencoding(F,K,fichout,met,ea1,solver) :
    print("initial formula :",F.toString())
    if (solver=='cqesto' or solver=='ghostq' or solver=='qfun') and (met != 'PNF1') and (met != 'PNF2') and (met != 'FBV') :
        PrintCol("Warning ! This solver requires a Prenex formula ! The reduction "+met+" does not ensure it !",'RED')
    if (solver=='cqesto' or solver=='ghostq' or solver=='qfun') and (ea1 == 2) :
        PrintCol("Warning ! This solver requires a Prenex formula and the encoding 2 for Ex1/Al1 does not ensure it.",'RED')
    (F,rp)=Reduction(F,met,solver)
    PNF=(met=='FBV' or met=='FFNN' or met=='PNF1' or met=='PNF2')  
    print("Rewriting of the formula: "+F.toString())
    BF=QBF(F,rp,K,met,PNF,ea1)
#    print('**',BF.toString())
    print("Taille de la formule QBF:",str(BF.Size()))
    if VERBOSE :
        print('**',BF.toString())
    if solver=='z3' :
        fichout=fichout+'.z3'
        toZ3(BF,fichout)
    else :
        fichout=fichout+'.qcir'
        toQCIR(BF,fichout)
    return fichout

#fichou est le nom de fichier ou sera stocke le code... un suffixe sera ajoute (soit .z3 soit .qcir)    
def Verif(F,K,fichout,met,ea1,solver) :
    global AUTOTEST
    fout = QBFencoding(F,K,fichout,met,ea1,solver)
    print("Solver...")
    if AUTOTEST :
        cmd="./"+solver+" "+fout+" > res_out"
        os.system(cmd)
        res= getResult(solver,'res_out')
        if res==None :
            PrintCol("Warning !",'MAGENTA')
            os.system('tail res_out')
        return res
    elif TIMEOUT :
        cmd="timeout 600 time ./"+solver+" "+fout
        vexit = os.system(cmd)
        print("val =",vexit)
        return None        
    else :
        cmd="time ./"+solver+" "+fout
        vexit = os.system(cmd)
        print("val =",vexit)
        return None

def getResult(solver,fichres) :
    fich=open(fichres,'r')
    l=fich.readlines()
    lastl=l[len(l)-1]
    if solver=='z3' :
        res=lastl[:-1]
        if res=='sat' :
            return True
        elif res=='unsat' :
            return False
        else :
            return None        
    elif solver=='qfm' or solver=='cqesto' :
        res=lastl[-2:-1]
        if res=='1' :
            return True
        elif res=='0' :
            return False
        else :
            return None
    
