# -----------------------------------------------------------------------------
# qctl.py
# Parser pour les formules QCTL
# -----------------------------------------------------------------------------

import sys
#sys.path.insert(0, "../..")

if sys.version_info[0] >= 3:
    raw_input = input

from  formule import *

reserved = { 'true':'TRUE', 'false':'FALSE', 'EF':'EF', 'AF':'AF', 'EG':'EG', 'AG':'AG', 'EX':'EX', 'AX':'AX', 'Ex1':'QEXISTS1', 'Ex': 'QEXISTS', 'Al1':'QFORALL1','Al':'QFORALL', 'E' : 'PEXISTS', 'A':'PFORALL','U':'UNTIL','W':'WEAKU'}

tokens = ['ID', 'NUMBER', 'EQUIV', 'IMPL']+list(reserved.values())

literals = ['=', '<', '>', '!', '&', '|', '(', ')','[',']',',','.']

# Tokens

t_EQUIV = r'<->'
t_IMPL = r'->'
#t_INF = r'<'
#t_SUP = r'>'

def t_ID(t) :
    r'[a-zA-Z][a-zA-Z0-9]*'
    t.type = reserved.get(t.value,'ID')
    return t


def t_NUMBER(t):
    r'\d+'
    t.value = int(t.value)
    return t

t_ignore = " \t"


def t_newline(t):
    r'\n+'
    t.lexer.lineno += t.value.count("\n")


def t_error(t):
    print("Illegal character '%s'" % t.value[0])
    t.lexer.skip(1)

# Build the lexer
import ply.lex as lex
lex.lex()

# Parsing rules

precedence = (
    ('right','QEXISTS', 'QEXISTS1', 'QFORALL', 'QFORALL1'), 
    ('left','EQUIV','IMPL'),
    ('left', '|'),
    ('left', '&'),
    ('right', '!', 'EF', 'AF', 'EG', 'AG', 'EX', 'AX', 'UNTIL', 'WEAKU'),
)

def p_statement_formula(p):
    'statement : formula'
#    print(p[1].toString())
    p[0]=p[1]


def p_formula_binop(p) :
    '''formula : formula '&' formula
                  | formula '|' formula
                  | formula EQUIV formula
                  | formula IMPL formula'''
    if p[2] == '&':
        p[0] = AND([p[1],p[3]])
    elif p[2] == '|':
        p[0] = OR([p[1],p[3]])
    elif p[2] == '<->':
        p[0] = EQ(p[1],p[3])
    elif p[2] == '->':
        p[0] = IMPL(p[1],p[3])

def p_formula_unop(p) :
    '''formula : EF formula
                  | AF formula
                  | EG formula
                  | AG formula
                  | EX formula
                  | AX formula
                  | '!' formula'''
    if p[1] == 'EF':
        p[0] = EF(p[2])
    elif p[1] == 'AF':
        p[0] = AF(p[2])
    elif p[1] == 'EG':
        p[0] = EG(p[2])
    elif p[1] == 'AG':
        p[0] = AG(p[2])
    elif p[1] == 'EX':
        p[0] = EX(p[2])
    elif p[1] == 'AX':
        p[0] = AX(p[2])
    elif p[1]=='!':
        p[0]= NOT(p[2])

def p_formula_EAUW(p) :
    '''formula : PEXISTS formula UNTIL formula
                  | PEXISTS formula WEAKU formula
                  | PFORALL formula UNTIL formula
                  | PFORALL formula WEAKU formula'''
    if p[1] == 'E' and p[3] == 'U':
        p[0] = EU(p[2],p[4])
    elif p[1] == 'E' and p[3] == 'W':
        p[0] = EW(p[2],p[4])
    elif p[1] == 'A' and p[3] == 'U':
        p[0] = AU(p[2],p[4])
    elif p[1] == 'A' and p[3] == 'W':
        p[0] = AW(p[2],p[4])

def p_formula_Q(p) :
    '''formula : QEXISTS lprop '.' formula 
                  | QFORALL lprop '.' formula
                  | QEXISTS1  lprop '.' formula
                  | QFORALL1 lprop '.' formula'''
    if p[1] == 'Ex' :
        p[0] = EXISTS(p[2],p[4])
    elif p[1] == 'Ex1' :
        p[0] = EXISTS1(p[2],p[4])
    elif p[1] == 'Al' :
        p[0] = FORALL(p[2],p[4])
    elif p[1] == 'Al1' :
        p[0] = FORALL1(p[2],p[4])
        
def p_formula_AP(p) :
    '''formula : ID'''
    p[0]=AP(p[1])

def p_formula_Bool(p) :
    '''formula : TRUE 
                  |  FALSE '''
    if p[1]=='true' :
        p[0]=TOP()
    elif p[1]=='false' :
        p[0]=BOTTOM()

    
# def p_lprop(p) :
#     '''lprop : ID
#                | lprop ',' ID'''
#     if (len(p) > 2) :
#         p[0] = p[1].append(p[3])
#     else :
#         p[0] = list(p[1])
                               
def p_lprop_ID(p) :
    '''lprop : ID'''
    p[0] = [p[1]]

def p_lprop_list(p) :
    '''lprop : lprop ',' ID'''
    L=p[1]
    L.append(p[3])
    p[0] = L[:]

def p_formula_group(p):
    '''formula : '(' formula ')' 
                | '[' formula ']' '''
    p[0] = p[2]


# def p_expression_number(p):
#     "expression : NUMBER"
#     p[0] = p[1]



def p_error(p):
    if p:
        print("Syntax error at '%s'" % p.value)
    else:
        print("Syntax error at EOF")

import ply.yacc as yacc
parser = yacc.yacc()

# while 1:
#     try:
#         s = raw_input('qctl > ')
#     except EOFError:
#         break
#     if not s:
#         continue
#     yacc.parse(s)
